#pragma once

#include <string>
#include <vector>
#include <array>
#include <memory>
#include <unordered_map>
#include <functional>

#include "tf_data_types.h"

namespace Trueface
{
    class SDK;

    /**
     * Class which is used to manage a face chip image.
     * A face chip represents and aligned and cropped face image.
     * A Facechip can be in CPU memory or GPU memory.
     */
    class Facechip {
    public:
        /**
         * Default constructor.
         */
        Facechip() = default;

        /**
         * Destructor frees allocated memory.
         */
        ~Facechip();
#if defined(TARGET_OS_IPHONE) || defined(TARGET_IPHONE_SIMULATOR)
        void destroy();
#endif

        /**
         * Save the face chip to disk.
         * @param filepath the filepath where the face chip should be saved, including the image extension.
         */
        void saveImage(const std::string& filepath);

        /**
         * Load the face chip from disk.
         * @param filepath the filepath of the facechip to load.
         * @param gpuMemory read the image into GPU memory. This should be set to true when running GPU inference.
         * @param gpuIndex the GPU index.
         */
        ErrorCode loadImage(const std::string& filepath, bool gpuMemory = false, int gpuIndex = 0);

        /**
         * Get the face chip height in pixels.
         * @return Returns the face chip height in pixels.
         */
        unsigned int getHeight();

        /**
         * Get the face chip width in pixels.
         * @return Returns the face chip width in pixels.
         */
        unsigned int getWidth();

        /**
        * Is the face chip in GPU memory.
        * @return Returns true if the face chip is in GPU memory.
        */
        bool isGPUImage();

        /**
        * Get the decoded image buffer.
        * @return Returns the decoded image buffer.
        */
        uint8_t* getData();
    private:
        unsigned int m_width    = 0;
        unsigned int m_height   = 0;
        bool m_GPUImage         = false;
        uint8_t* m_data         = nullptr;
        friend class SDK;
    };

     /*
     * Shared pointer of type Facechip.
     */
    using TFFacechip = std::shared_ptr<Facechip>;

    /**
     * Class which is used to manage a decoded image which is in the format expected by inference functions.
     * Should populate Image using the preprocessImage functions.
     * An Image can be in CPU memory or GPU memory.
     */
    class Image {
    public:
        /**
         * Default constructor.
         */
        Image() = default;
        /**
         * Destructor frees allocated memory.
         */
        ~Image();
#if defined(TARGET_OS_IPHONE) || defined(TARGET_IPHONE_SIMULATOR)
        void destroy();
#endif

        /**
         * Rotate the image.
         *
         * @param rotateFlag the RotateFlags.
         */
        void rotate(RotateFlags rotateFlags);

        /**
         * Resize the image to the target size.
         * Adds padding to the bottom and right if required in order to maintain aspect ratio.
         * @param width the target width.
         * @param height the target height.
         */
        void resize(unsigned int width, unsigned int height);

        /**
         * Save the preprocessed image to disk.
         * @param filepath the filepath where the image should be saved, including the image extension.
         */
        void saveImage(const std::string& filepath);

        /**
         * Get the image height in pixels.
         * @return Returns the image height in pixels.
         */
        unsigned int getHeight() const;

        /**
         * Get the image width in pixels.
         * @return Returns the image width in pixels.
         */
        unsigned int  getWidth() const;

        /**
         * Get the number of channels for the image.
         * @return Returns the number of channels.
         */
        unsigned int getChannels() const;

        /**
         * Get the GPU stride for GPU images.
         * @return Returns the GPU stride for the image.
         */
        unsigned int getStride() const;

        /**
        * Is the image in in GPU memory.
        * @return Returns true if the image is in GPU memory.
        */
        bool isGPUImage() const;

        /**
         * Is the memory managed by the class.
         * If the memory is managed by the class, the image data will be freed when the instance storage duration ends.
         * If the memory is not managed by the class, then it is up to the user to free the allocated memory.
         * @return Returns true if the memory is managed by the class.
         */
        bool isManagedBuffer() const;

        /**
        * Get the decoded image buffer.
        * @return Returns the decoded image buffer.
        */
        uint8_t* getData() const;
    private:
        unsigned int m_height   = 0;
        unsigned int m_width    = 0;
        unsigned int m_channels = 0;
        unsigned int m_stride   = 0;
        bool m_GPUImage         = false;
        bool m_managedBuffer    = true;
        uint8_t* m_data         = nullptr;
        friend class SDK;
        friend class Impl;
    };

    /*
     * Shared pointer of type Image.
     */
    using TFImage = std::shared_ptr<Image>;

    class SDK {
        public:
            /**
             * Initialize the SDK using default \ref ConfigurationOptions.
             *
             * @sa ConfigurationOptions
             */
            SDK();

            /**
             * Do not allow copy.
             */
            SDK(const SDK& other) = delete;

            /**
             * Do not allow assignment operator.
             */
            SDK& operator=(const SDK& other) = delete;

            /**
             * Move constructor.
             * If other is to be used after move, must be re-initialized.
             */
            SDK(SDK&& other) noexcept;

            /**
             * Move assignment operator.
             * If other is to be used after move, must be re-initialized.
             */
            SDK& operator=(SDK&& other) noexcept;

            /**
             * Initialize the SDK using custom \ref ConfigurationOptions.
             *
             * @param options custom configuration options.
             */
            SDK(const ConfigurationOptions& options);

            ~SDK();

            /**
             * Uninitialize specific module.
             *
             * @param uninitializeModule module to clear.
             */
            void uninitializeModule(UninitializeModule uninitializeModule);

            /**
             *  Sets and validates the given license token. Need to call this method
             *  before being able to use the SDK.
             *
             * @param[in] token the license token (if you do not have this talk to
             *        support@trueface.ai).
             * @return Whether the given license token is valid.
             * 
             * @sa isLicensed()
            */
            bool setLicense(std::string token);
#if defined(TARGET_OS_IPHONE) || defined(TARGET_IPHONE_SIMULATOR)
            bool setLicense(std::string token, std::string package_id);
#endif
            /** 
             * Checks whether the given license token is valid and you can use
             * the SDK.
             *
             * @param[out] daysRemaining the number of remaining days that the token is valid for.
             * @return Whether the given license token is valid.
             * 
             * @sa setLicense()
            */
            bool isLicensed(unsigned int& daysRemaining);

            /** 
             * Gets the version-build number of the SDK.
             *
             *
             * @return Version Number as a string.
            */
            static std::string getVersion();

            /**
             * Obtain the unique fingerprint for the current device.
             * The fingerprint is used to generate a hardware locked token.
             *
             * @return The fingerprint as a string.
             */

            static std::string getDeviceFingerprint();

            /**
             * Obtain the GPU UUID(s).
             * The GPU UUID(s) are used to generate a GPU locked token.
             * Note, this function can only be run on the GPU SDK, otherwise it will throw an exception.
             *
             * @return The GPU UUID(s).
             */
            static std::string getGPUUUID();

            /**
             *  Preprocess the image to be used by the other methods.
             *  Creates a copy of the image buffer.
             *
             * @param[in] image an 8-bit decoded image array, in the CPU memory or the GPU memory.
             * @param[in] width the image width.
             * @param[in] height the image height.
             * @param[in] color the image color model, see ColorCode.
             * @param[out] tfImage the preprocesses image.
             * @param[in] stride the distance between image array rows in bytes, also known as step size (used only for pointers to the GPU memory).
             *
             * @return error code, see ErrorCode.
             *
            */
            ErrorCode preprocessImage(uint8_t* image, uint16_t width, uint16_t height, ColorCode color, TFImage& tfImage, int stride=0);

            /**
             * Preprocess the image to be used by other methods.
             * Buffer is not copied and memory should be managed by the caller.
             * More efficient than preprocessImage() as the data is referenced and not copied.
             * Buffer must remain valid until the TFImage is no longer needed, at which point the caller is responsible for freeing the allocated memory.
             * Note: If a CPU memory image is provided when using the GPU SDK, the SDK will copy the image to GPU memory, and thus the image will not be referenced.
             * To optimize performance with the GPU SDK, ensure the image is already loaded into GPU memory before calling this method.
             * @param image an 8-bit decoded image array in RGB format, in the CPU memory or the GPU memory.
             * @param width the image width.
             * @param height the image height.
             * @param tfImage the preprocessed image.
             * @param stride the distance between image array rows in bytes, also known as step size (used only for pointers to the GPU memory).
             * @return
             */
            ErrorCode preprocessRgbImage(uint8_t* image, uint16_t width, uint16_t height, TFImage& tfImage, int stride = 0);

            /**
             * Preprocess the image to be used by the other methods.
             *
             * @param[in] encodedImageBuffer encoded image buffer, supports JPG, PNG, BMP, and TIFF.
             * @param[out] tfImage the preprocesses image.
             *
             * @return error code, see ErrorCode.
             *
            */
            ErrorCode preprocessImage(const std::vector<uint8_t>& encodedImageBuffer, TFImage& tfImage);

            /** 
             * Preprocess the image to be used by the other methods.
             *
             * @param[in] imageFile the path of a JPEG, JPG, PNG, BMP, or TIFF file (path must not contain ~ ).
             * @param[out] tfImage the preprocesses image.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode preprocessImage(const std::string& imageFile, TFImage& tfImage);

            /**
             * Detect people and objects in the image.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[out] boundingBoxes a vector of BoundingBox objects that
             *             gets populated with the detected objects.
             *
             * @return error code, see ErrorCode.
             *
             * @sa getObjectLabelString(ObjectLabel label)
            */
            ErrorCode detectObjects(const TFImage& tfImage, std::vector<BoundingBox>& boundingBoxes);

            /**
             * Draw object detector bounding boxes and labels on the image.
             * Note, you should not run any inference functions on the image after calling this method.
             *
             * @param [in] tfImage the input image on which to draw the labels.
             * @param [in] boundingBoxes the bounding boxes returned by detectObjects().
             * @return error code, see ErrorCode.
             */
            ErrorCode drawObjectLabels(const TFImage& tfImage, const std::vector<BoundingBox>& boundingBoxes);

            /** 
             * Estimate body pose.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] boundingBoxes bounding boxes returned by detectObjects().
             * @param[out] landmarks a vector of Landmark objects that
             *             represent body joints.
             * @return error code, see ErrorCode.
            */
            ErrorCode estimatePose(const TFImage& tfImage, const std::vector<BoundingBox>& boundingBoxes, std::vector<std::vector<Landmark>>& landmarks);

            /** 
             * Draw body pose on the image.
             * Note, you should not run any inference functions on the image after calling this method.
             *
             * @param[in] tfImage the input image on which to draw the pose.
             * @param[in] landmarks a vector of Landmark objects that
             *             represent body joints.
             * @return error code, see ErrorCode.
            */

            ErrorCode drawPose(const TFImage& tfImage, const std::vector<std::vector<Landmark>>& landmarks);

            /**
             *  Converts an object label to a string. This can be used to
             *  display the object labels.
             * 
             * @param[in] label the label of a BoundingBox.
             * @return The object label as a string.
            */
            static std::string getObjectLabelString(ObjectLabel label);

            /**
             * Detect the largest face in the image.
             * This method has a small false positive rate.
             * To reduce the false positive rate to near zero, filter out faces with score lower than 0.90.
             * Alternatively, you can use the \ref ConfigurationOptions.fdFilter
             * configuration option to filter the detected faces.
             * See detectFaces() for the detection scale range.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[out] faceBoxAndLandmarks the FaceBoxAndLandmarks containing the landmarks and bounding box
             * of the largest detected face in the image.
             * @param[out] found whether a face was found in the image.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode detectLargestFace(const TFImage& tfImage, FaceBoxAndLandmarks& faceBoxAndLandmarks, bool& found);

            /**
             * Detect all the faces in the image. This method has a small false positive rate.
             * To reduce the false positive rate to near zero, filter out faces with score lower than 0.90.
             * Alternatively, you can use the \ref ConfigurationOptions.fdFilter
             * configuration option to filter the detected faces.
             *
             * The face detector has a detection scale range of about 5 octaves. \ref ConfigurationOptions.smallestFaceHeight
             * determines the lower of the detection scale range. E.g., setting \ref ConfigurationOptions.smallestFaceHeight to
             * 40 pixels yields the detection scale range of ~40 pixels to 1280 (=40x2^5) pixels.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[out] faceBoxAndLandmarks a vector of \ref FaceBoxAndLandmarks representing each of the detected faces.
             * If not faces are found, the vector will be empty. The detected faces are sorted in order of descending face score.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode detectFaces(const TFImage& tfImage, std::vector<FaceBoxAndLandmarks>& faceBoxAndLandmarks);

            /**
             * Obtain the 106 face landmarks.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] landmarks an array of 106 face landmark points.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getFaceLandmarks(const TFImage& tfImage, const FaceBoxAndLandmarks& faceBoxAndLandmarks, Landmarks& landmarks);

            /**
             *  Align the the detected face to be optimized for passing to
             *  feature extraction. If using the face chip with Trueface algorithms (ex face recognition),
             *  do not change the default margin and scale values.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] faceBoxAndLandmarks the FaceBoxAndLandmarks returned
             *            by detectLargestFace() or detectFaces().
             * @param[out] tfFacechip A TFFacechip to store the output facechip.
             * @param[in] marginLeft adds a margin to the left side of the face chip.
             * @param[in] marginTop adds a margin to the top side of the face chip.
             * @param[in] marginRight adds a margin to the right side of the face chip.
             * @param[in] marginBottom adds a margin to the bottom side of the face chip.
             * @param[in] scale changes the scale of the face chip.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode extractAlignedFace(const TFImage& tfImage,
                                         const FaceBoxAndLandmarks& faceBoxAndLandmarks,
                                         TFFacechip& tfFacechip,
                                         int marginLeft=0,
                                         int marginTop=0,
                                         int marginRight=0,
                                         int marginBottom=0,
                                         float scale=1.0);

            /**
             * Detect the largest face in the image and return its feature
             * vector.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[out] faceprint a Faceprint object which will contain
             *             the face feature vector.
             * @param[out] foundFace indicates if a face was detected in the image.
             * If no face was detected, then the faceprint will be empty.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getLargestFaceFeatureVector(const TFImage& tfImage, Faceprint &faceprint, bool& foundFace);

            /**
             * Extract the face feature vector from an aligned face chip.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] tfFacechip TFFacechip returned by extractAlignedFace().
             * Face image must be have size of 112x112 pixels (default extractAlignedFace() margin and scale values).
             * @param[out] faceprint a Faceprint object which will contain
             *             the face feature vector.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getFaceFeatureVector(const TFFacechip& tfFacechip, Faceprint& faceprint);


            /**
             * Extract the face feature vectors from the face chips.
             * This batch processing method increases throughput when using GPU inference.
             *
             * @param[in] tfFacechips vector of TFFacechip.
             * @param[out] faceprints vector of Faceprint object which will contain
             *             the face feature vectors.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getFaceFeatureVectors(const std::vector<TFFacechip>& tfFacechips, std::vector<Faceprint>& faceprints);

            /**
             * Extract the face feature vector from the face box.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] faceBoxAndLandmarks the FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] faceprint Faceprint to contain the face template.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getFaceFeatureVector(const TFImage& tfImage, const FaceBoxAndLandmarks& faceBoxAndLandmarks, Faceprint& faceprint);

            /**
             * Convert a Faceprint into a json string.
             *
             * @param[in] faceprint The Faceprint to stringify.
             *
             * @return The json string representation of the Faceprint.
            */
            static std::string faceprintToJson(const Faceprint& faceprint);

            /**
             * Populate a Faceprint from a json string.
             *
             * @param[in] jsonStr The json string representation of a Faceprint, generated from the faceprintToJson() function.
             * @param[out] faceprint The Faceprint which will be populated from the json string.
             *
             * @return error code, see ErrorCode.
            */
            static ErrorCode jsonToFaceprint(const std::string& jsonStr, Faceprint& faceprint);

            /**
             * Compute the similarity between two feature vectors, or how
             * similar two faces are.
             *
             * @param[in] faceprint1 the first Faceprint to be compared.
             * @param[in] faceprint2 the second Faceprint to be compared.
             * @param[out] matchProbability the probability the two face feature vectors are a match.
             * @param[out] similarityMeasure the computed similarity measure.
             *
             * @return error code, see ErrorCode.
             *
             * @sa getLargestFaceFeatureVector
            */
            static ErrorCode getSimilarity(const Faceprint& faceprint1,
                                    const Faceprint& faceprint2,
                                    float &matchProbability,
                                    float &similarityMeasure);

            /**
             * Estimate the head pose.
             * Only works with images where the face is at least 100 pixels in height, otherwise returns `ErrorCode::FACE_TOO_SMALL`.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] yaw the rotation angle around the image's vertical axis, in radians.
             * @param[out] pitch the rotation angle around the image's transverse axis, in radians.
             * @param[out] roll the rotation angle around the image's longitudinal axis, in radians.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode estimateHeadOrientation(const TFImage& tfImage, const FaceBoxAndLandmarks& faceBoxAndLandmarks, float& yaw, float& pitch, float& roll);

            /**
             * Estimate the quality of the face image for recognition.
             *
             * @param[in] tfFacechip The TFFacechip returned by extractAlignedFace().
             * @param[out] quality a value between 0 to 1, 1 being prefect quality for recognition.
             * We suggest using a threshold of 0.999 as a filter for enrollment images.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode estimateFaceImageQuality(const TFFacechip& tfFacechip, float& quality);

            /**
             * Determine if the image is over or under exposed, indicating suboptimal lighting conditions for face recognition.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] faceBoxAndLandmarks the FaceBoxAndLandmarks of the face to analyze.
             *
             * @return error code, see ErrorCode.
             * If the image has correct exposure, will return ErrorCode::NO_ERROR.
             */
            ErrorCode checkFaceImageExposure(const TFImage& tfImage, const FaceBoxAndLandmarks& faceBoxAndLandmarks);

            /**
             * Determine if the eyes of the specified face are open or closed.
             * The default thresholds for this function have been selected to minimize
             * False Positives (an open eye classified as a blink) at the expensive of
             * increased False Negatives (a closed eye classified as an open eye).
             * However, you can also choose your own thresholds based on the members of `Trueface::BlinkState`.
             * Note, this function does not work well with eyeglasses, which can be detected using `Trueface::SDK::detectGlasses`.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] targetFace the face on which to run blink detection.
             * @param[out] blinkState the output of the blink prediction.
             *
             * @return error code, see ErrorCode. Note, if the face is not frontal facing, an error will be returned.
            */
            ErrorCode detectBlink(const TFImage& tfImage, const FaceBoxAndLandmarks& targetFace, BlinkState& blinkState);

            /**
             * Ensures that the face size meets the requirements for active spoof.
             * Must check return value of function!
             * Active spoof works by analyzing the way a persons face changes as they move closer to a camera.
             * The active spoof solution therefore expects the face a certain distance from the camera.
             * **In the far image, the face should be about 18 inches from the camera, while in the near image,
             * the face should be 7-8 inches from the camera.**
             * This function must be called before calling detectActiveSpoof().
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] faceBoxAndLandmarks The face on which to run active spoof detection.
             * @param[in] activeSpoofStage The stage of the image, either near stage or far stage.

             * @return error code, see ErrorCode.
             * If `ErrorCode::NO_ERROR` is returned, then the image is eligible for active spoof detection.
             * If `ErrorCode::FACE_TOO_CLOSE` or `ErrorCode::FACE_TOO_FAR` is returned, the image is not eligible for active spoof detection.
             *
             */
            ErrorCode checkSpoofImageFaceSize(const TFImage& tfImage, const FaceBoxAndLandmarks& faceBoxAndLandmarks,
                                              ActiveSpoofStage activeSpoofStage);


            /**
             *
             * Detect if there is a presentation attack attempt.
             * Must call checkSpoofImageFaceSize() on both input faces before calling this function.
             *
             * @param nearFaceLandmarks The face landmarks of the near face, obtained by calling getFaceLandmarks().
             * @param farFaceLandmarks The face landmarks of the far face, obtained by calling getFaceLandmarks().
             * @param spoofScore The output spoof score.
             * If the spoof score is above the threshold, then it is classified as a real face.
             * If the spoof score is below the threshold, then it is classified as a fake face.
             * @param spoofPrediction The predicted spoof result, using a spoofScore threshold of 1.05.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode detectActiveSpoof(const Landmarks& nearFaceLandmarks, const Landmarks& farFaceLandmarks,
                                        float& spoofScore, SpoofLabel& spoofPrediction);

            /**
             * Detect if there is a presentation attack attempt.
             * Passive spoof detection only works with "selfie-style", meaning the face must be up close and take up most of the frame.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] result the SpoofLabel results for the face image.
             * @param[out] spoofScore the probability that the image is a spoof attempt
             * (1 indicates real image, 0 indicates spoof attempt).
             * @param[in] threshold the spoof score threshold above which it is considered a real image.
             * Refer to our <a href="https://docs.trueface.ai/roc-curves">ROC curves</a> when selecting a threshold.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode detectSpoof(const TFImage& tfImage, const FaceBoxAndLandmarks& faceBoxAndLandmarks,
                                  SpoofLabel& result, float& spoofScore, float threshold = 0.75);


            /**
             *
             * Create a connection to a new or existing database.
             * If the database does not exist, a new one will be created with the provided name.
             * If the `Trueface::DatabaseManagementSystem::NONE` (memory only)
             * configuration option is selected, this function does not need to be called (and is a harmless no-op).
             * If connecting to a hosted PostgreSQL database (AWS, Digital Ocean, etc), see the `HOSTED_DATABASE` environment variable
             * on the <a href="./environment.html">environment variable page</a>.
             * If connecting to a PostgreSQL database hosted on the same machine, you must specify the "host" parameter.
             * If connecting to a PostgreSQL database hosted on a remote server or a different machine on the same network,
             * must specify the "hostaddr" parameter instead with the URL or IP of said database.
             *
             * @param databaseConnectionString
             * If `Trueface::DatabaseManagementSystem::SQLITE` is selected, this should be the filepath to the database.
             * ex. "/myPath/myDatabase.db".
             * If `Trueface::DatabaseManagementSystem::POSTGRESQL`
             * is selected, this should be a database connection string.
             * <a href="https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS">Here</a>
             * is a list of all supported PostgreSQL connection parameters.
             * ex. "hostaddr=192.168.1.0 port=5432 dbname=face_recognition user=postgres password=my_password".
             * ex. "host=localhost port=5432 dbname=face_recognition user=postgres password=m_password".
             * To enable ssl, add "sslmode=require" to the connection string.
             *
             * @return error code, see ErrorCode.
             */

            ErrorCode createDatabaseConnection(const std::string& databaseConnectionString);

            /**
             *
             * Create a new collection, or load data from an existing collection into memory (RAM) if one with the
             * provided name already exists in the database.
             * Equivalent to calling createCollection() then loadCollection().
             *
             * @param[in] collectionName the name of the collection.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode createLoadCollection(const std::string& collectionName);

            /**
             *
             * Create a new collection in the database.
             * Before enrolling Faceprints into newly created collection, must call loadCollection().
             * If the collection with the provided name already exists, this is a harmless no-op.
             *
             * @param[in] collectionName the name of the collection.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode createCollection(const std::string& collectionName);

            /**
             *
             * Loads the collection into memory.
             * Must be called before enrolling Faceprints or calling identification functions.
             *
             * @param[in] collectionName the name of the collection to load into memory.
             *
             * @return error code, see ErrorCode.
             * Will return an error if the collection does not exist.
             */
            ErrorCode loadCollection(const std::string& collectionName);

            /**
             *
             * Deletes a collection from the current database.
             *
             * @param[in] collectionName the name of the collection to delete.
             *
             * @return error code, see ErrorCode.
             * Will return an error if the collection does not exist.
             */
            ErrorCode deleteCollection(const std::string& collectionName);

            /**
             * Enroll a Faceprint for a new or existing identity in the collection.
             *
             * @param[in] faceprint the Faceprint to enroll in the collection.
             * @param[in] identity the identity corresponding to the Faceprint.
             * @param[out] UUID universally unique identifier corresponding to the Faceprint.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode enrollFaceprint(const Faceprint& faceprint, const std::string& identity, std::string& UUID);

            /**
             * Get a list of the names of all the collections in the database.
             * Collection names can then be passed to getCollectionMetadata() and getCollectionIdentities().
             *
             * @param[out] collectionNames list of collection names in the database.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode getCollectionNames(std::vector<std::string>& collectionNames);

            /**
             * Get the metadata for the specified collection in the database, loaded or unloaded.
             *
             * @param[in] collectionName the name of the collection for which to retrieve the metadata.
             * @param[out] metadata the metadata for the specified collection.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode getCollectionMetadata(const std::string& collectionName, CollectionMetadata& metadata);

            /**
             * Get a map of identities and UUIDs for the specified collection in the database, loaded or unloaded.
             * This can be a slow operation (especially for unloaded collections), call sparingly.
             *
             * @param[in] collectionName the name of the collection for which to retrieve the identities.
             * @param[out] identities a map of identities and corresponding UUIDs.
             * There can be multiple UUIDs mapped to a single identity if multiple Faceprints were enrolled for that identity.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode getCollectionIdentities(const std::string& collectionName,
                                                  std::unordered_multimap<std::string, std::string>& identities);

            /**
             * Remove a Faceprint from the collection using the UUID.
             *
             * @param[in] UUID the universally unique identifier returned by enrollFaceprint().
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode removeByUUID(const std::string& UUID);

            /**
             * Remove all Faceprints in the collection corresponding to the identity.
             *
             * @param[in] identity the identity to remove from the collection.
             * @param[out] numFaceprintsRemoved the the number of Faceprint which were removed for that identity.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode removeByIdentity(const std::string& identity, unsigned int& numFaceprintsRemoved);

            /**
             * Get the top match Candidate in the collection and the corresponding similarity score and match probability.
             *
             * @param[in] faceprint the Faceprint to be identified.
             * @param[out] candidate the top match Candidate.
             * @param[out] found set to true if a match is found.
             * @param[in] threshold the similarity score threshold above which it is considered a match.
             * Higher thresholds may result in faster queries.
             * Refer to our <a href="https://docs.trueface.ai/roc-curves">ROC curves</a> when selecting a threshold.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode identifyTopCandidate(const Faceprint& faceprint, Candidate& candidate, bool& found, float threshold = 0.4f);

            /**
             * Get the top match Candidate in the collection and the corresponding similarity score and match probability
             * for each probe Faceprint. Like identifyTopCandidate(), but runs search queries in parallel and improves throughput.
             *
             * @param[in] faceprints a vector of Faceprint to be identified.
             * @param[out] candidates the top match Candidate for each probe Faceprint.
             * @param[out] found set to true if a match is found for the probe Faceprint.
             * @param[in] threshold the similarity score threshold above which it is considered a match.
             * Higher thresholds may result in faster queries.
             * Refer to out <a href="https://docs.trueface.ai/roc-curves">ROC curves</a> when selecting a threshold.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode batchIdentifyTopCandidate(const std::vector<Faceprint>& faceprints, std::vector<Candidate>& candidates,
                    std::vector<bool>& found, float threshold = 0.4f);

            /**
             * Get a list of the top n match Candidate in the collection and their corresponding similarity scores and match probabilities.
             * Ideal for investigation mode.
             *
             * @param[in] faceprint the Faceprint to be identified.
             * @param[out] candidates a list of the top n matches (or fewer) Candidate, ordered by descending similarity score.
             * @param[out] found set to true if at leaset one match is found.
             * @param[in] numCandidates max number of Candidate identities to return.
             * @param[in] threshold the similarity score threshold above which it is considered a match.
             * Higher thresholds may result in faster queries.
             * Refer to out <a href="https://docs.trueface.ai/roc-curves">ROC curves</a> when selecting a threshold.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode identifyTopCandidates(const Faceprint& faceprint, std::vector<Candidate>& candidates, bool& found,
                            unsigned int numCandidates, float threshold = 0.4f);

            /**
             * Detect whether the face in the image is wearing a mask or not.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] result The predicted MaskLabel for face image.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode detectMask(const TFImage& tfImage, const FaceBoxAndLandmarks& faceBoxAndLandmarks, MaskLabel& result);

            /**
            * Detect whether the faces are wearing a mask or not.
            * This batch processing method increases throughput when using GPU inference.
            *
            * @param[in] chips A vector of TFFacechip on which to run mask detection.
            * @param[out] results The predicted MaskLabel for the input face chips.
            *
            * @return error code, see ErrorCode.
            */
            ErrorCode detectMasks(const std::vector<TFFacechip>& chips, std::vector<MaskLabel>& results);

            /**
             * Detect whether the face in the image is wearing any type of eye glasses or not.
             *
             * @param[in] tfImage the input image returned by preprocessImage().
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] result The predicted GlassesLabel for face image.
             * @param[out] glassesScore The glasses score for this image. This can be used for setting custom thresholds that work
             * better for the use case. By default, we use a glasses score greater than 0.0 to determine that glasses were detected.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode detectGlasses(const TFImage& tfImage, const FaceBoxAndLandmarks& faceBoxAndLandmarks, GlassesLabel& result, float& glassesScore);

    private:
            class Impl;
            std::unique_ptr<Impl> pImpl;
            friend class Internal;
            friend class Image;
    };

} // namespace truface
